//////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2022 zSpace, Inc.  All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////

#ifndef __ZSPACE_CORE_COMPATIBILITY_H__
#define __ZSPACE_CORE_COMPATIBILITY_H__

#include "zSpaceCoreCompatibilityTypes.h"
#include "zSpaceCoreCompatibilityEntryPointFuncTypes.h"
#include "zSpaceCoreCompatibilityReflection.h"

#endif // __ZSPACE_CORE_COMPATIBILITY_H__
