// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef vtk_exodusII_h
#define vtk_exodusII_h

/* Use the exodusII library configured for VTK.  */
#include <vtkexodusII/include/exodusII.h>

#endif
