// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#include "windows.h"
#include <mfidl.h> // must be included first

#include <mfapi.h>
#include <mferror.h>
#include <mfreadwrite.h>

int main()
{
  return 0;
}
