basisGradient[0] = 0.5 - rr;
basisGradient[1] = 0.;
basisGradient[2] = 0.;

basisGradient[3] = 0.5 + rr;
basisGradient[4] = 0.;
basisGradient[5] = 0.;

basisGradient[3] = 2.0;
basisGradient[4] = 0.;
basisGradient[5] = 0.;
