// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
/**
 * @class   vtkHDFVersion
 * @brief   version number for legacy readers/writers
 */

#ifndef vtkHDFVersion_h
#define vtkHDFVersion_h

#include "vtkABINamespace.h"

VTK_ABI_NAMESPACE_BEGIN
constexpr int vtkHDFMajorVersion = 2;
constexpr int vtkHDFMinorVersion = 5;

VTK_ABI_NAMESPACE_END
#endif // vtkHDFVersion_h
// VTK-HeaderTest-Exclude: vtkHDFVersion.h
