// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-FileCopyrightText: Copyright (c) 2021, Triad National Security, LLC
// SPDX-License-Identifier: LicenseRef-BSD-3-Clause-LANL-Triad-USGov

#include <PIOData.h>

VTK_ABI_NAMESPACE_BEGIN
PIO_DATA::PIO_DATA() = default;

PIO_DATA::~PIO_DATA() = default;
VTK_ABI_NAMESPACE_END
