// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause

#define vtkImageProgressIterator_cxx

#include "vtkImageProgressIterator.txx"

VTK_ABI_NAMESPACE_BEGIN
vtkInstantiateTemplateMacro(template class VTKCOMMONEXECUTIONMODEL_EXPORT vtkImageProgressIterator);
VTK_ABI_NAMESPACE_END
